#include "tree.H"
#include <stdio.h>
#include "sse.H"
#include "debug.H"
#include "analysis.H"
#include <math.h>
#include "vec_utility.H"
#include <fstream>
using std::ofstream;
#include "predict.H"
#include "protein.H"
#include "matrix.H"
#include <utility>
using std::pair;
#include "ioRoutines.H"

/**@name analysis.C */
//@{
extern int makeLinkBetweenNodes( int numNodes, const int n1, const int n2, 
				 const float forwardPathWeight, 
				 const float backwardPathWeight);

extern int mycompare       (const void *vfirst, const void *vsecond);

extern void printMatrix( mat4 m);
extern void printFloatArray(float * A, int numElements);

extern int floatCompare (const void* pv1, const void* pv2);

float rmsd          (const  Protein * p1, const Protein * p2);

int visitsClass::numNonZeroElements = 0;

void initNodeColor(const int numNodes, int thecolor, Protein ** allNodes) {

  int i;

  for (i = 0; i < numNodes; i++) {
    allNodes[i]->myNode->nodeColor = (char)(thecolor);
    allNodes[i]->myNode->nodeTime = -1;
  }

}

void modifiedInitNodeColor(int numberOfNodes, Protein ** allNodes) {

  int i;

  for (i = 0; i < numberOfNodes; i++) {
    if (allNodes[i]->myNode->nodeColor != 3) 
      allNodes[i]->myNode->nodeColor = 0;
    allNodes[i]->myNode->nodeTime = -1;
  }
}

void correctRange(int * a, int lowerBound, int upperBound) {
    if ((*a) < lowerBound) 
	(*a) = lowerBound;
    if ((*a) > upperBound)
	(*a) = upperBound;
}

bool checkRange(float f, float lowerBound, float upperBound) {
    if ((f > upperBound) || (f < lowerBound))
	return false;
    else return true;
}

int inverse_sort_compare(const void *vfirst, const void *vsecond) {

  visitsClass *first;
  visitsClass *second;
  
  first  = (visitsClass*)(vfirst);
  second = (visitsClass*)(vsecond);

  if (first->numVisited < second->numVisited) {
    return 1;
  }
  else if (first->numVisited == second->numVisited) {
    return 0;
  }
  else {
    return -1;
  }
}



// returns avg. energy, also computes and prints min, max energies.
void findAvgEnergy (Protein ** allNodes, int numNodes, float * avgEnergy)
{

  const float maxFloat = 50000.0;
  float minEnergy = maxFloat; 
  float maxEnergy = -maxFloat;
  float energy    =  0.0;

  (*avgEnergy) =  0.0;
  for (int i = 0; i < numNodes; i++) 
    { 
      energy = allNodes[i]->getEnergy();
      // min energy also includes native structure!!
      if (energy < minEnergy) 
	{
	  minEnergy = energy;
	}
      if (energy > maxEnergy) 
	{
	  maxEnergy = energy;
	}
      (*avgEnergy) += energy;
    }

  (*avgEnergy) = (*avgEnergy)/((float)(numNodes));
  
  cout << "Min Node Energy = " << minEnergy << endl;
  cout << "Max Node Energy = " << maxEnergy << endl;
  cout << "Avg Node Energy = " << *avgEnergy << endl;
}




///returns true if i connected to j, false otherwise.
bool connected (int i, int j, Protein ** allNodes, int numNodes)
{
  ptrList * linkPtr;
  bool retval = false;
  linkPtr     = allNodes[i]->myNode->firstLink;

  assert ((i < numNodes) && (i >= 0));
  while (linkPtr != NULL) 
    {
      int index = linkPtr->node->myConfig->getIndex();
      if (index == j)
	{
	  retval = true;
	  break;
	}
      linkPtr       = linkPtr->nextPtr;
    }

  return retval;
}

///returns num neighbors of node # i.
unsigned int findNumNeighbors (int i, int numNodes, Protein ** allNodes)
{
  ptrList * linkPtr;
  unsigned int retval = 0;
  linkPtr  = allNodes[i]->myNode->firstLink;

  assert ((i < numNodes) && (i >= 0));
  while (linkPtr != NULL) 
    {
      //int index = linkPtr->node->myConfig->getIndex();
      retval++;
      /*
	I don't see the point of the below check, so commenting it out.
	if (index > numNodes)
	{
	cout  << "index = " << index << " numnodes = " << numNodes << endl;
	}
	assert         (index < numNodes);
      */
      linkPtr       = linkPtr->nextPtr;
    }

  return retval;
}


///computes pfold, writes the result to a file.
void computeAndPrintPfold (vector<Real> & pfolds, int numPfoldsRequested, int maxNumNeighbors, 
			   int runIndex,
			   float foldedMacrostateRMSDRadius, 
			   float unfoldedMacrostateRMSDRadius, 
			   int numberOfNodes,
			   Protein**  allNodes,
			   int foldedStateIndex, int unfoldedStateIndex)
{ 

  vector<int> foldedIndices, unfoldedIndices;
  //cerr << "analysis.C: (using " << ((sizeof(Real)==sizeof(float))? "float": "double") << ")" << endl;
  
  foldedIndices   = computeCloseNodes(foldedStateIndex, allNodes, 
				      numberOfNodes,
				      foldedMacrostateRMSDRadius);

  unfoldedIndices = computeCloseNodes(unfoldedStateIndex, 
				      allNodes, 
				      numberOfNodes,
				      unfoldedMacrostateRMSDRadius);
  
  //  pfolds = computePfoldWrapper( numberOfNodes, allNodes, 
  computePfoldWrapper( pfolds, numberOfNodes, allNodes, 
		       numPfoldsRequested, 
		       foldedIndices,
		       unfoldedIndices,
		       maxNumNeighbors);
  
  /*
  cout << "analysis.C" << endl;
  for (unsigned int i = 0; i < pfolds.size(); i++)
    {
      cout << pfolds[i] << endl;
    }
  */
  if (pfolds.size() != (unsigned int) numPfoldsRequested)
    cout << "pfold size = " << pfolds.size() << " numPfoldsRequested = " << numPfoldsRequested << endl;
  assert (pfolds.size() == (unsigned int) numPfoldsRequested);
  printPfolds    (pfolds, numberOfNodes,  runIndex);
}

///
vector<int> computeCloseNodes(int centerNodeIndex, Protein ** allNodes, 
			      int numNodes,
			      float macrostateRMSDRadius)
{
  vector<int> retval(0);

  assert ((centerNodeIndex >= 0) && (centerNodeIndex < numNodes));
  
  for (int i = 0; i < numNodes; i++)
    {
      if (allNodes[centerNodeIndex]->lsqfit_RMSD(allNodes[i]) < macrostateRMSDRadius)
	{
	  retval.push_back(i);
	}
    }
  return retval;
}

#if 0
///returns the sorted order of variances, in descending order. Only returns the first sizeOrder of them.
void sortFloatVector(vector<int> & order, const vector<float> & variances, int sizeOrder)
{
  static bool firstCall = true;
  FILE * fout;
  
  if (firstCall)
    {
      fout = fopen("variances.txt", "w");
      cout << "check variances.txt" << endl;
      firstCall = false;
    }
  else
    {
      fout = fopen("variances.txt", "a");
    }
  
  pair<float,int>* distArray = new pair<float,int>[variances.size()];
  
  order.resize(sizeOrder);
  
  for (unsigned int i=0; i<variances.size(); i++)
    {
      assert (variances[i] >= 0.0);
      distArray[i].first  = -variances[i];
      distArray[i].second = i;
    }

  qsort(distArray, variances.size(), sizeof(pair<float,int>), floatCompare);

  for (int i=0;i<sizeOrder;i++)
    {
      order[i]  = distArray[i].second;
      assert ((order[i] >= 0) && ((unsigned int) order[i] < variances.size()));
      printf("%d %f\n", order[i], distArray[i].first);
    }
  //  fprintf(fout, "%d %f\n", order[0], distArray[0].first);
	
  fclose(fout);
  delete[] distArray;
}
#endif
//@}

