#include "ioRoutines.H"
#include "tree.H"
#include <stdio.h>
#include <stdlib.h>
#include <fstream>
#include "predict.H"
#include "sse.H"
#include <math.h>
#include <iostream>
#include <sys/times.h>
#include <unistd.h>

using std::cerr;
using std::endl;
using std::cout;

#define SUCCESS 1
#define FAILURE 0

/**@name ioRoutines.C */
//@{
class edge
{
public:
  int otherNodeIndex;
  float pW1, pW2; 
  /*
    pW1 is from the node on which this edge is to node #
    otherNodeIndex. pW2 is the other way around. otherNodeIndex
    should be > index of the node on which the edge is defined.
  */
};

class connection 
{
public:
  vector<edge> edges;
};

vector<connection> connections;
char transitionProbsFilename[100];
extern bool writeTransitionProbs;

void extStr(char* out, const char* in, int st, int end) {
 
  int outcount = 0; 
  int i;

  for (i = st - 1; i < end; i++) { 
 
    out[outcount] = in[i]; 
    outcount++; 
  } 
 
  out[outcount] = '\0'; 
 
}

/* 
   reads the nodes, for the vectorbased representation. 
   the format of the file is :
   <num Nodes>
   for each node :
     for each sse vector:
       dihedral, vector, twist angles, length of the protein.
     energy of that node.
   numNodes at the beginning should be 1, and allNodes[0]
   is assumed to already contain the folded state.
*/

void readNodes       (Protein ** allNodes, int * numNodes,
		      const char filename[] ) 
{
  nodeClass * newNode;
  FILE * fin = fopen(filename,"r");
  int         numSseVectors, numNodes2Read;
  
  assert (*numNodes == 1);

  if (fin == NULL)
    {
      cerr << filename << " was not opened. returning." << endl;
    }

  fscanf (fin, "%d\n",  &numNodes2Read);

  numSseVectors = Protein::getNumSseElements();

  float         _dih, _vec, _twi, _len;
  float         energy;
  vector<float> dih, vec, twi, len;
  
  dih.resize(numSseVectors); 
  vec.resize(numSseVectors); 
  twi.resize(numSseVectors); 
  len.resize(numSseVectors); 

  int nodeIndex = 1;

  while (!feof(fin))
    {
      
      allNodes[nodeIndex] = new Protein(nodeIndex);

      for (int i = 0; i < numSseVectors; i++) 
	{
	  //	  cerr << " i = " << i << endl;
	  fscanf(fin, "%f %f %f %f\n", &_dih, &_vec, &_twi, &_len);
	  
	  dih[i] = _dih;  vec[i] = _vec;	  
	  twi[i] = _twi;  len[i] = _len;
	}
      
      fscanf(fin, "%f\n", &energy);

      allNodes[nodeIndex]->createGivenConfig  (dih, vec, twi,len,
					       allNodes[0]);
      
      if (fabs(allNodes[nodeIndex]->getEnergy() - energy) >= 1E-3)
	{
	  cout << "read energy = " << energy << " computed : " << allNodes[*numNodes]->getEnergy() << endl;
	  assert (0);
	}
      
      newNode                 = new nodeClass;
      newNode->myConfig       = allNodes[nodeIndex];
      allNodes[nodeIndex]->setMyNode(newNode);
      allNodes[nodeIndex]->checkBondDihedralAngles();
      nodeIndex ++;
    }
  fclose(fin);
  
  *numNodes = nodeIndex;
  assert ((*numNodes) == numNodes2Read + 1);
  cerr << "read " << numNodes2Read << " nodes." << endl;
}

void readConnections (const char filename[], int numNodes)
{
 FILE * fin;
 
 int i, j;
 float neighbourDist,forwardPathWeight,  backwardPathWeight;
 

  fin = fopen(filename, "r");
  if (fin == NULL) 
    {
      cerr << filename << " was not opened. exiting." << endl;
      exit(0);
    }
  
  while (!feof(fin))
    {
      fscanf (fin, "%d %d %f %f %f \n",  &i, &j, &neighbourDist, 
	      &forwardPathWeight,  &backwardPathWeight);
      makeLinkBetweenNodes (numNodes, i,j,forwardPathWeight, 
			    backwardPathWeight);
    }
  fclose(fin);

}

/* 
   just creates the nodes, does not set the energy or anything. 
   startIndex is the node index from which we start allocating numNodes.
   startIndex is generally 0, but is useful if the folded state is already 
   allocated at node #0, in which case numNodes parameter is actually 1 less 
   than the number of nodes we want to create.
*/

void allocateNodes(int numNodes, Protein ** allNodes, int startIndex)
{
  int finalNodeIndex = startIndex + numNodes;
  for (int i = startIndex; i < finalNodeIndex; i++)
    {
      allNodes[i]                         = new Protein(i);
      nodeClass * newNode                 = new nodeClass;
      newNode->myConfig                   = allNodes[i];
      allNodes[i]->setMyNode(newNode);
    }
}

/*
  reads one line, and stores it in allNodes[*nodeIndex].
  the line format is described in readMyRoadmap.
  updates connections array.

  also, if WRITE_TRANSITION_PROB is set, writes the 
  transition probabilities, including the prob. of 
  staying at the current node. the format is :
  n1 n2 pTrans.
*/

int readOneLineOfMyRoadmap (FILE * fin, int * nodeIndex, 
			    Protein ** allNodes
			    //vector<int> &   connectedNode1, 
			    //vector<int> &   connectedNode2,
			    //vector<float>&  edgeWeight
			    )
{
  static bool   firstCall = true;
  static int    numSSEs;
  float         _dih, _vec, _twi, _len,energy;
  static        vector<float> dih, vec, twi, len;
  int           numEdges;

  if (firstCall)
    {
      firstCall = false;
      numSSEs   = Protein::getNumSseElements();
      dih.resize(numSSEs); 
      vec.resize(numSSEs); 
      twi.resize(numSSEs); 
      len.resize(numSSEs);
    }
  
  fscanf(fin, "%d", nodeIndex);
  
  // in case eof not encountered yet..
  
  if (feof(fin))
    return FAILURE;
  
  for (int j = 0; j < numSSEs; j++)
    {
      fscanf (fin, "%f %f %f %f",  &_dih, &_vec, &_twi, &_len);
      dih[j] = _dih;  vec[j] = _vec;	  
      twi[j] = _twi;  len[j] = _len;
    }
      
  if ((*nodeIndex) != 0)
    {
      allNodes[(*nodeIndex)]->createGivenConfig  (dih, vec, 
						  twi, len,
						  allNodes[0]);
      allNodes[(*nodeIndex)]->computeEnergy();
      allNodes[(*nodeIndex)]->findParams();
    }
     
  int dummyInt1, dummyInt2; 
  fscanf  (fin, "%d %d", &dummyInt1, &dummyInt2);
  assert  ((dummyInt1 == -1) && (dummyInt2 == -1));
  
  fscanf  (fin, "%f %d", &energy, &numEdges);
  
  if (fabs(allNodes[(*nodeIndex)]->getEnergy() - energy) >= 0.3)
    {
      cout << "read energy = " << energy << " computed : " << allNodes[(*nodeIndex)]->getEnergy() << endl;
      cout << "node index  = " << (*nodeIndex) << endl;
      assert (0);
    }
  else 
    {
      //      cout << "for node # " << *nodeIndex << " energy difference = " << fabs(allNodes[(*nodeIndex)]->getEnergy() - energy) << endl;
    }
  
  vector<int> adjacentNodeList;
  float pTrans;
  vector<float> pTransArray;

  adjacentNodeList.resize(0);
  pTransArray.resize(0);

  // now reading the adjacency list
  for (int k = 0; k < numEdges; k++)
    {
      int   otherNodeIndex;
      float pW;
      int   dummyInt;
	  
      fscanf         (fin, "%d %d %f", &otherNodeIndex,&dummyInt, &pW);
      
      assert         (dummyInt == 1);
      
      adjacentNodeList.push_back(otherNodeIndex);
     
      
      insertConnection(*nodeIndex, otherNodeIndex, pW);
      
      pTrans                                        = exp(-pW);
      pTransArray.push_back(pTrans);
      
    }


  if (writeTransitionProbs)
    {
      assert (strcmp(transitionProbsFilename, "") != 0);
      cout << "check : " << transitionProbsFilename << endl;
      FILE * fout = fopen(transitionProbsFilename, "a");
      assert (fout != NULL);

      for (unsigned int i = 0; i < pTransArray.size(); i++)
	{
	  fprintf(fout, "%d %d %f\n", (*nodeIndex), 
		  adjacentNodeList[i],
		  pTransArray [i]);

	  cout << "printed " << (*nodeIndex) << " " << adjacentNodeList[i] << " ";
	  cout << pTransArray [i] << endl;
	}
      fclose(fout);

    }
  return SUCCESS;
}

/*
  reads the roadmap file, which has the following format:
  
  the map format is as follows:

  minEnergy
  samples_per_region
  samples_around_goal
  numRegionsSampledAround (including the folded state)
  
  on the fifth line:
    (# of nodes) 

  then followed by the detail information for each node, the format is:

  (index) (dihedral, vector, twist angles, length) for each sse; -1 -1 (potential) (# of nodes connecting to this node) (the adjacent list) 

    the adjacent list format is:  (neighour node index) 1 (edge weight)
    
  establishes the nodes and the connections.

  the first node is supposed to be already read and allocated.
 
  (index = 0..numNodes - 1).   

  establishes the nodes and the connections.

  if writeTransitionProbs is set, writes transition probabilities.
*/

void readMyRoadmap    (int * numRandomNodes, int * numberOfNodes, 
		       const char * filename, 
		       Protein ** allNodes, int runIndex)
{

  FILE * fin =  fopen (filename,"r");
  int    numNodes, numSSEs,  nodeIndex;


  //vector<int> connectedNode1, connectedNode2;
  //vector<float> edgeWeight;
  // there is an edge between connectedNode1[i] and connectedNode2[i]
  // of path weight edgeWeight[i].
  /*
    connectedNode1.resize(0);
    connectedNode2.resize(0);
    edgeWeight.resize(0);
  */
  
  assert (*numberOfNodes == 1);
  
  numSSEs = Protein::getNumSseElements();
  
  if (fin == NULL)
    {
      cout << "couldn't open " << filename << ". exiting." << endl;
      exit(0);
    }
  else
    {
      cout << "reading " << filename << endl;
    }
  
  // discard the first line for the time being
  for (int i = 0; i < 1; i++)
    {
      char   line[200];
      fgets (line, 200, fin);
    }
  int samples_per_region, samples_around_goal, numRegionsSampledAround;

  fscanf(fin, "%d\n", &samples_per_region     );
  fscanf(fin, "%d\n", &samples_around_goal    );
  fscanf(fin, "%d\n", &numRegionsSampledAround);
  fscanf(fin, "%d\n", &numNodes);

  //#if WRITE_TRANSITION_PROB  
  FILE * fout;
  if (writeTransitionProbs)
    {
      sprintf (transitionProbsFilename, "transitionProbs%d_%d.txt", numNodes, 
	       runIndex);
      fout   = fopen(transitionProbsFilename, "w");  
      fclose(fout);
      cout << "check " << transitionProbsFilename << endl;
    }
  //#endif
  
  if (numRegionsSampledAround > 0) 
    {
      // this number includes samples 
      // both around the folded state and the other low energy configs.
      *numRandomNodes = numNodes - 
	(numRegionsSampledAround - 1) * samples_per_region - 
	samples_around_goal;
    }
  else
    {
      *numRandomNodes = numNodes;
    }
  
  connections.resize(numNodes);

  allocateNodes     (numNodes - 1, allNodes, 1); // skipping the first protein

  for (int i = 0; i < numNodes; i++)
    {
      addNode2Forest(i);
    }

  int i = 0; // just to check the indices of the array.

  while (!feof(fin))
    {
      if (!readOneLineOfMyRoadmap (fin, &nodeIndex, allNodes)) 
	//	connectedNode1, connectedNode2, edgeWeight))
	break;
      //cout << *allNodes[nodeIndex] << endl;
      assert (nodeIndex == i);
      i++;
    }
  
  
  if (numNodes != i)
    {
      cerr << "alert!. # nodes read = " << nodeIndex << " whereas the file is supposed to contain "   << numNodes  << " nodes." << endl;
      assert (0);
    }

  /*  
      #if WRITE_TRANSITION_PROB
      vector<int> indices;
      unsigned int  K = 10;
  
      findMinEnergyClusters    (indices,  K, 
      allNodes,   numNodes - 1000,
      5.0, 0);

  
      fout = fopen("transitionProbs.txt", "a");

      for (int i  = 0; i < K; i++)
      {
      fprintf(fout, "%d\n", indices[i]);
      }

      fclose(fout);
      //  cout << "wrote transition probabilities.. exiting." << endl;
      cout << "wrote edge weights and the lowest 10 energy configs." << endl;
      #endif
  */
  //  cout << "total " << connectedNode1.size() << " connections to be made." << endl;
  //doConnections4MyRoadmap(connectedNode1, connectedNode2, edgeWeight, numNodes);

  //cout << "skipping doing connections..." << endl;
  doConnections4MyRoadmap  (numNodes);
  
  fclose(fin);
  (*numberOfNodes) = numNodes;
  cout << "read a roadmap of " << numNodes << " nodes." << endl;
}


void writeTransitionProbabilities(int numNodes, Protein ** allNodes, const char * filename)
{
  //char filename[] = "transitionProbs.txt";
  FILE * fout     =  fopen (filename,"w");
  vector<float> tranProbs;
  vector<int>   neighborIndices;
  
  cout << "check " << filename << endl;
  for (int i = 0; i < numNodes; i++)
    {
      findNeighborsAndTransitionProbabilities (neighborIndices,
					       tranProbs,
					       i, 
					       allNodes);
      assert (neighborIndices.size() == tranProbs.size());

      for (unsigned int j = 0; j < neighborIndices.size() ; j++)
	{
	  assert (tranProbs[j] >= 0);
	  assert (neighborIndices[j] != i);
	  fprintf(fout, "%d %d %e \n", i, neighborIndices[j], tranProbs[j]);
	}
    }
  fclose(fout);
}

/*
  same format as in readMyRoadmap.
*/

void writeMyRoadmap    (int numNodes, const char * filename, 
			Protein ** allNodes, int SAMPLES_PER_REGION,
			int SAMPLES_AROUND_GOAL, int MIN_REGIONS)

{
  FILE * fout =  fopen (filename,"w");
  vector<float> edgeWeights;
  vector<int>   neighborIndices;
  int numSSEs     = Protein::getNumSseElements();
  
  cout << "check " << filename << endl;

  fprintf(fout, "%f\n", Protein::minEnergy);
  fprintf(fout, "%d\n", SAMPLES_PER_REGION);
  fprintf(fout, "%d\n", SAMPLES_AROUND_GOAL);
  fprintf(fout, "%d\n", MIN_REGIONS);

  fprintf(fout, "%d\n", numNodes);

  for (int i = 0; i < numNodes; i++)
    {
      fprintf(fout, "%d ", i);
      for (int j = 0; j < numSSEs; j++)
	{
	  Sse * ptr2Sse = allNodes[i]->getSseElement(j);
	  fprintf(fout, "%f %f %f %f ", 
		  ptr2Sse->getDihedralAngle(),
		  ptr2Sse->getVectorAngle(),
		  ptr2Sse->getTwistAngle(),
		  ptr2Sse->getLength());
	}

      fprintf(fout, "-1 -1 %f ", allNodes[i]->getEnergy());

      findNeighborsAndEdgeWeights (neighborIndices,
				   edgeWeights,
				   i, 
				   allNodes);
      
      int numNeighbors = neighborIndices.size();
      
      fprintf(fout, "%d ", numNeighbors);
      
      for (int j = 0; j <  numNeighbors; j++)
	{
	  fprintf(fout, "%d 1 %f ", neighborIndices[j], edgeWeights[j]);
	}
      fprintf(fout, "\n");
    }
  fclose(fout);
}

/*
  inserts the edge of path weight pW from n1 to n2 to the connection array.
*/
void insertConnection(int n1, int n2, float pW)
{
  const int UNINITIALIZED = -1;
  edge  myEdge;

  if (n1 < n2)
    {
      myEdge.otherNodeIndex = n2;
      myEdge.pW1            = pW;
      myEdge.pW2            = UNINITIALIZED;
      connections[n1].edges.push_back(myEdge);
    } 
  else
    {
      assert (n1 > n2);
      // n1 > n2, 
      // first find the edge from n2 to n1.
      // then set pW2 parameter to pW.
      // due to roadmap writing scheme, the edge from n2 to n1 should already
      // be written. 

      bool found = false;
      unsigned int i;
      // becomes true when we find part of the connection between n1 and n2
      // that should aalready have been stored.


      for (i = 0; i < connections[n2].edges.size(); i++)
	{
	  if (connections[n2].edges[i].otherNodeIndex == n1)
	    {
	      found = true;
	      break;
	    }
	}
      
      assert (found);
      assert (connections[n2].edges[i].pW2 == UNINITIALIZED);
      connections[n2].edges[i].pW2 = pW;
    }
}

/*
  uses connections array do the connections.
*/
void doConnections4MyRoadmap(unsigned int numNodes)
{
  assert (numNodes == connections.size());

  for (unsigned int i = 0; i < numNodes; i++)
    {
      for (unsigned int j = 0; j < connections[i].edges.size(); j++)
	{
	  edge myEdge = connections[i].edges[j];
	  makeLinkBetweenNodes (numNodes, i, myEdge.otherNodeIndex, 
			        myEdge.pW1, myEdge.pW2);
	  /*
	  joinTheTreesForNodes (i, 
				myEdge.otherNodeIndex,
				numNodes);
	  */
	}
    }
}

void printEnergies(Protein ** allNodes, int numberOfNodes)
{
  FILE * fout = fopen("energies.txt", "w");
  cout << "check energies.txt" << endl;
  for (int i = 0; i < numberOfNodes; i++)
    {
      fprintf(fout, "%f\n", allNodes[i]->getEnergy());
    }
  fclose(fout);
}           

void printRMSDs(Protein ** allNodes, int numberOfNodes)
{
  FILE * fout = fopen("rmsds.txt", "w");
  cout << "check rmsds.txt" << endl;
  for (int i = 0; i < numberOfNodes; i++)
    {
      fprintf(fout, "%f\n", allNodes[0]->lsqfit_RMSD(allNodes[i]));
    }
  fclose(fout);
}       


/*
    given a roadmap file, goes thru the unnecessary info
    and picks the nodes with indices inputStartIndex..inputEndIndex. 
    Adds them to the 
    roadmap, starting from node # startNodeIndex. 
    Assumes the same format as in readMyRoadmap.
    updates  numNodes.

    inputStartIndex..inputEndIndex have to match the index in the input roadmap
    file in order to be included in the new roadmap.
    (Since the roadmap file numbers the roadmap indices with 0-basis,
     this means that a inputStartIndex of 0 is acceptable.) 
*/

void readSomeNodesFromFile(const char *  roadmapFilename,
			   Protein  **   allNodes, 
			   int        *  numNodes, 
			   int inputStartIndex, int inputEndIndex)
{
  int           numEdges, startNodeIndex, dummyInt1, dummyInt2, numSSEs;
  char          temp[1500];
  float         energy, _dih, _vec, _twi, _len;
  vector<int>   indices;
  vector<float> dih, vec, twi, len;

  indices.resize(0);


  for (int i = inputStartIndex; i <= inputEndIndex; i++)
    {
      indices.push_back(i);
    }
  
  FILE * fin =  fopen (roadmapFilename,"r");  
  if (fin == NULL)
    {
      cout << "couldn't open " << roadmapFilename << ". exiting." << endl;
      exit(0);
    }
  else
    {
      cout << "reading " << roadmapFilename << endl;
    }
  
  startNodeIndex =  *numNodes;

  assert          (startNodeIndex >= 0);
  
  allocateNodes   (indices.size(), allNodes, startNodeIndex);
  
  // discard the first line for the time being
  for (int i = 0; i < 1; i++)
    {
      char   line[200];
      fgets (line, 200, fin);
    }

  int samples_per_region, samples_around_goal, numRegionsSampledAround;

  fscanf(fin, "%d\n", &samples_per_region     );
  fscanf(fin, "%d\n", &samples_around_goal    );
  fscanf(fin, "%d\n", &numRegionsSampledAround);
  fscanf(fin, "%d\n", &dummyInt1); // numnodes in the roadmap read. 

  int prevNodeIndex = -1, nodeIndex;
  unsigned int numNodesRead = 0;
  numSSEs = Protein::getNumSseElements();
  dih.resize(numSSEs);  vec.resize(numSSEs);
  twi.resize(numSSEs);  len.resize(numSSEs);

  while (numNodesRead < indices.size())
    {
      fscanf(fin, "%d", &nodeIndex);
      // cout << "nodeIndex = " << nodeIndex << endl;
      assert (prevNodeIndex + 1 == nodeIndex);
      prevNodeIndex ++;
      
      if (feof(fin))
	{
	  cout << "node info missing..." << endl;
	  exit(0);
	}

      if (nodeIndex == indices[numNodesRead])
	{
	  int newNodeIndex                 = numNodesRead + startNodeIndex;

	  for (int j = 0; j < numSSEs; j++)
	    {
	      fscanf (fin, "%f %f %f %f",  &_dih, &_vec, &_twi, &_len);
	      dih[j] = _dih;  vec[j] = _vec;	  
	      twi[j] = _twi;  len[j] = _len;
	    }
	  
	  allNodes[newNodeIndex]->createGivenConfig  (dih, vec, twi, len,
						      allNodes[0]);
	  allNodes[newNodeIndex]->computeEnergy();
	  allNodes[newNodeIndex]->findParams();
	  

	  fscanf  (fin, "%d %d", &dummyInt1, &dummyInt2);
	  assert   ((dummyInt1 == -1) && (dummyInt2 == -1));
	  
	  //  read energy;
	  fscanf   (fin, "%f %d", &energy, &numEdges);

	  //cout << "reading node # " << nodeIndex << " info. energy = ";
	  //cout << energy << endl;
	  
	  //  check energy;

	  if (fabs(allNodes[newNodeIndex]->getEnergy() - energy) >= 0.5)
	    {
	      cout << "read energy = " << energy << " computed : " << allNodes[newNodeIndex]->getEnergy() << endl;
	      cout << "node index  = " << newNodeIndex  << endl;
	      assert (0);
	    }

	  fgets(temp, 1500, fin);
	  //cout << "discarding the following conn. info : " << temp << endl;
	  numNodesRead ++;
	}
      else
	{
	  fgets(temp, 1500, fin);
	  //cout << "discarding the following complete node info : " << temp << endl;
	}
    }
  
  *numNodes += numNodesRead;
  cout << "read " << indices.size() << " nodes." << endl;
  cout << "num nodes at the end of the function : " << *numNodes << endl; 
  fclose(fin);
}


float  sumFloatArray(float * array, int numElements) {

  float retval = 0.0;
  int   i;

  for (i = 0; i < numElements; i++) {
    retval += array[i];
  }
  return retval;
}

// filename = "tests/" + prot_name + fileNameSuffix;
void   prepareFileName(char * filename, 
		       const char * prot_name, const char *fileNameSuffix) {

  strcpy(filename, "tests/");
  strcat(filename, prot_name);
  strcat(filename, "/");
  strcat(filename, fileNameSuffix);
}

void printFloatArray(float * A, int numElements) {
  int i;

  for (i = 0;  i<  numElements; i++) {
    cout << A[i] << " ";
  }
  cout << endl;
}


void printIntArray(int * A, int numElements) 
{
  int i;

  cout << "content of the array : ";
  for (i = 0;  i<  numElements; i++) {
    cout << A[i] << " ";
  }
  cout << endl;

}

void printMatrix(mat4 m)
{
   int i,j;

  for (i = 0; i< 4; i++) 
    {
      for (j = 0; j < 4; j ++) 
	{
	  printf ("%f  ", m[i][j]);
	}
      printf("\n");
    }

  printf("\n");  
}


/*
  prints the indices of nodes that are close to the folded or unfolded nodes
  in a file, in 1-based format.
*/
void printCloseNodeIndices (int foldedStateIndex, int unfoldedStateIndex,
			    float minRMSDFolded, float minRMSDUnfolded, 
			    Protein ** allNodes, int numNodes,
			    int runIndex)
{ 
  
  vector<int>     close2FoldedNodeIndices, close2UnfoldedNodeIndices;
  char filename[100];
  
  close2FoldedNodeIndices  .resize(0);
  close2UnfoldedNodeIndices.resize(0);

  cout << "numNodes = " << numNodes << endl;
  
  for ( int i = 0; i < numNodes; i++)
    {
      if (allNodes[foldedStateIndex]->lsqfit_RMSD(allNodes[i]) < minRMSDFolded)
	{
	  close2FoldedNodeIndices.push_back(i+1);
	}
      if (allNodes[unfoldedStateIndex]->lsqfit_RMSD(allNodes[i]) < minRMSDUnfolded)
	{
	  close2UnfoldedNodeIndices.push_back(i+1);
	}
    }
  
  sprintf        (filename, "close2Folded%d_%d.txt", numNodes, runIndex);
  cout << "check " << filename << endl;
  ofstream  fout2(filename);

  fout2 << "% numnodes = " << numNodes << " runIndex = " << runIndex << endl; 
  fout2 << "% radius of the folded macrostate : " << minRMSDFolded << endl;
  fout2 << "% node indices (1-based) of nodes close 2 folded state nodes are : " << endl;

  for ( unsigned int i = 0; i < close2FoldedNodeIndices.size(); i++)
    {
      fout2 << close2FoldedNodeIndices[i] << " "; 
    }
  fout2 << endl;
  fout2.close();

  sprintf        (filename, "close2Unfolded%d_%d.txt", numNodes, runIndex);

  fout2.open(filename);
  cout << "check " << filename << endl;
  fout2 << "% numnodes = " << numNodes << " runIndex = " << runIndex << endl; 
  fout2 << "% radius of the unfolded macrostate : " << minRMSDUnfolded << endl;
  fout2 << "% node indices (1-based) of close 2 unfolded state nodes are : " << endl;

  for ( unsigned int i = 0; i < close2UnfoldedNodeIndices.size(); i++)
    {
      fout2 << close2UnfoldedNodeIndices[i] << " " ;
    }
  fout2 << endl;
  fout2.close();
}

void writeConnection(int i, int j, const string&  connectionFilename, 
		     float neighbourDist, float forwardPathWeight, 
		     float backwardPathWeight)
{
  FILE * fout;
  
  fout = fopen(connectionFilename.c_str(), "a");
  
  fprintf(fout, "%d %d %f %f %f \n",  i, j, neighbourDist, forwardPathWeight, 
	  backwardPathWeight);
  fclose(fout);
}





/*
  startIndex is an addition that tells from which roadmap node
  to start MC simulation.

  runIndex is the suffix that is to be added to the output filenames.

  maxFoldedRMSDRadius  : max distance (RMSD wise) of nodes from folded state
  maxUnfoldedRMSDRadius: max distance (RMSD wise) of nodes from unfolded state

*/
void handleInput (int argc, char ** argv, int * numRandomNodes,
		  char * relDir, char * protein_name,
		  //		  char * roadmapFilename, int *startIndex)
		  char * roadmapFilename, int *startIndex, int * runIndex,
		  bool * createNewRoadmap, 
		  float * maxFoldedRMSDRadius,
		  float * maxUnfoldedRMSDRadius)
{

  //cout << "also input the DOF range of parameters;" << endl;

  if (argc < 9) 
    {
      cout << "Usage: " << argv[0] << " <protein_name> <roadmap file_name> <create_new_nodes?<Y/N> > <write_transition_probs? <Y/N> > <startIndex>  <runIndex> <maxFoldedRMSD> <maxUnfoldedRMSD> [ init num nodes ]    " << endl;

      exit(0);
    }
  else
    {
      strcpy       (protein_name,    argv[1]);
      strcpy       (roadmapFilename, argv[2]);
      strcpy       (relDir, "tests/");
      strcat       (relDir, argv[1]);
      strcat       (relDir, "/");

      *startIndex    = (int)(atoi(argv[5]));
      //      cout << "start index = " << *startIndex << endl;

      *runIndex      = (int)(atoi(argv[6]));
      //cout << "run   index = " << *runIndex << endl;

      (* maxFoldedRMSDRadius)     = (float) (atof(argv[7]));
      (* maxUnfoldedRMSDRadius)   = (float) (atof(argv[8]));

      // get write_transition_probs
      switch(toupper(argv[4][0]))
	{
	case 'Y' : 
	  {
	    writeTransitionProbs = true;
	    //cout << "will write transition probabilities.." << endl;
	    break;
	  }
	case 'N' : 
	  {
	    writeTransitionProbs = false;
	    break;
	  }
	default : 
	  {
	    cout << " error! enter Y or N for write transition probs..." << endl;
	    assert (0);
	  }                  
	} // switch

      if (argc == 9)
	{
	  assert       ((argv[3][0] == 'n') || (argv[3][0] == 'N'));   
	  *createNewRoadmap = false; 
	  *numRandomNodes  = 0;
	}
      else 
	{
	  // argc > 9
	  // get num random nodes
	  assert              (argc > 9);

	  //	  cout << "will start from node # " << *startIndex << " to MC simulation." << endl;

	  assert              (toupper(argv[3][0]) == 'Y');
	  *createNewRoadmap  = true; 
	  (*numRandomNodes) = (int)(atoi(argv[9]));
	}
    }
}

// prints what are the parameters in the current run. should be called after
// handle input.
void printParameters(int  numRandomNodes,       const char * protName,  
		     bool createNewRoadmap)
{
  cout << "protein name             : " << protName       << endl;

  if (createNewRoadmap)
    {
      cout << "num. Random Nodes        : " << numRandomNodes << endl;
      cout << "Protein minEnergy set to : " << Protein::minEnergy << endl;
    }
}



void initialMessages()
{
  //  bool MCSim = false;
#if MC_SIMULATE
  cout << "MC simulation mode" << endl;
  MCSim = true;
#endif
#if _ANALYZE_PATH_WEIGHTS
  cout << "will analyze path weights." << endl;
#endif
#if WRITE_TO_FILE
  cout << "will write 2 file." << endl;
#endif

#if ANALYZE_ORDER
  assert (!MCSim); // MC. sim. should be off in order to run 
  // best paths from differing startins positions.
  cout << "analyzing order of coming together of secondary structure elements." << endl;
#endif

#if GAUSSIAN_SAMPLING
  cout << "doing gaussian sampling" << endl;
#else
  cout << "if creating a roadmap, doing normal sampling + NO extra sampling at regions of low energy" << endl;
#endif
}

void extract2 (Point point, const char * line) 
{
  char record[10];
  
  extStr(record,line,11,16);
  point[0] = (float) (atof(record));
  extStr(record,line,17, 23);
  point[1] = (float) (atof(record));
  extStr(record,line,24,31);
  point[2] = (float) (atof(record));

 
}


void 
reportTimes (struct tms start, struct tms end)
{
  float ticks = sysconf(_SC_CLK_TCK);//CLK_TCK;
  float utime = (end.tms_utime - start.tms_utime) / ticks;
  float stime = (end.tms_stime - start.tms_stime) / ticks;
  cout << "\t User time\t= " << utime << endl;
  cout << "\t System time\t= " << stime << endl;
  cout << "\t Total time\t= " << utime + stime << "\n\n";
  
}
// the ordered energies...
void printEnergiesLine(ofstream * outFile, int node1, int node2, 
		       const vector<float> & newOrderedE)
{
  // cout<<node1<<endl;
  // cout<<node2<<endl;
  (*outFile)<<node1<<" ";
  (*outFile)<<node2<<" ";
  (*outFile)<<newOrderedE.size()<<" ";
  for (unsigned int p = 0; p < newOrderedE.size(); p++)
    {
      // cout<<OrderedE[p]<<" ";
      (*outFile) << newOrderedE[p] << " ";
    }
  //(*outFile)<<"2001" << endl; 
  (*outFile)<<endl; 
}


void initializePCNNFile(const char * pcnnResultsFilename, FILE **fin, 
			int numberOfNodes, 
			int maxNumLinks)
{
  (*  fin) = fopen(pcnnResultsFilename, "r");
  int k, tmp, numNodes;
  
  if ((*fin) == NULL)
    {
      cout << "couldn't read " << pcnnResultsFilename << ". exiting." << endl;
      exit(0);
    }
  else
    cout << "reading : " << pcnnResultsFilename << endl;
  
  fscanf (*fin, "%d %d %d %d", &numNodes, &k, &tmp, &tmp);
  assert (numNodes == numberOfNodes);
  assert (k        == Protein::NUMDOF); // can remove later
  if (k < maxNumLinks)
    {
      cout << "warning.. The number of nearest neighbors returned from pcnn is not enough.";
      cout << endl << "Increase the number of nearest neighbors to be returned by pcnn or ";
      cout << endl << " decrease the number of connections to be made per node before ";
      cout << "running." << endl;
      cout << "Exiting..." << endl;
      exit(0);
    }
}


/// prints cA coordinates to a file, in PCNN format. 
void printPCNNFile(int numNodes, Protein ** allNodes, const char * filename)
{
  ofstream   fout (filename);
  
  cout << "check out " <<  filename  << endl;
  fout << numNodes << " " << Protein::numCalphas << " 1" << endl;
  fout.close();

  for (int i = 0; i < numNodes; i++)
    {
      allNodes[i]->appendCNNFile(filename);
    }
}


/**
   assumes that fin points to a new line of fin, with numLinks numbers  (in 1-based) corresponding to the nearest neighbors of currentNodeIndex. 

   currentNodeIndex is 0-based.

*/
void readOneLineOfPCNNFile(vector<int> & neighbors, FILE * fin, int numNodes, 
			   int currentNodeIndex, int numLinks)
{
  int neighborIndex;
  
  assert (fin != NULL);
  neighbors.resize(numLinks);

  assert ((currentNodeIndex >= 0) && (currentNodeIndex < numNodes));
  //cout << "node " << currentNodeIndex << " have the following nearest neighbors: " << endl;
  for (int j = 0; j < numLinks; j++)
    {
      fscanf (fin, "%d", &neighborIndex);
      //cout << neighborIndex - 1 << " ";
      assert ((neighborIndex >= 1) && (neighborIndex <= numNodes));
      assert (neighborIndex != currentNodeIndex + 1); //just making sure that pcnn does not return the node itself as one of the nearest neighbors.
      neighbors[j] = neighborIndex - 1; 
    }
  //cout << endl;
}

///
void linkNodesStatPrintOuts(int numNodes, int numTrees, 
			    Protein ** allNodes, treeClass *globalFirstTree)
{

  cout << "Nodes Done: " << numNodes << " Num Trees: " << numTrees << endl;
 
  //cout << "Trees created " << tempCreated << " trees removed " << tempDestroyed << endl;
  if (globalFirstTree != NULL) 
    globalFirstTree->printTree();

  if (allNodes[0]->myTree != NULL) 
    {
      cout << "numNodes in the tree of native structure : " << allNodes[0]->myTree->numNodesInTree << endl;
    }
 
  if (allNodes[1]->myTree != NULL) 
    {
      cout << "numNodes in the tree of node #1 : " << allNodes[1]->myTree->numNodesInTree << endl;
    }
  if (numTrees == 1)
    assert (allNodes[0]->myTree->numNodesInTree == numNodes);
}


void printNodeEnergies (Protein ** allNodes, int numberOfNodes)
{
  ofstream fout ("nodeEnergies.txt");
  cout << " check nodeEnergies.txt" << endl;
  
  for (int i = 0; i < numberOfNodes; i++)
    {
      fout << allNodes[i]->getEnergy() << endl;
    }
}


///normally this function should be in analysis.C but for some reason it does not link there.
void printPfolds    (const vector<Real>  & pfolds, int numNodes, int runIndex)
{
  char pfoldResultFilename[1000];

  sprintf(pfoldResultFilename, "pfold%d_%d.txt", numNodes, runIndex);
  ofstream fout(pfoldResultFilename);
  
  cout << "check " << pfoldResultFilename << endl;
  
  for (unsigned int i = 0; i < pfolds.size(); i++)
    {
      fout << pfolds[i] << endl;
    }


}


//@}
