// for matrix operations


#include "matrix.H"
#include <set>
using std::set;
#include <vector>
using std::vector;
#include <iostream>
using std::cout;
using std::endl;
#include "tree.H"


#define	errmesg(mesg)	printf("Error: %s error: line %d\n",mesg,__LINE__);

#ifndef __cplusplus
this should cause error if __cplusplus is not defined.
#endif



// extern is to tell the compiler that the function has C linkage and calling conventions.
extern "C"  SPMAT	*spLUfactor(SPMAT	*A, PERM	*px,
				    double alpha);

extern "C" VEC	        *spLUsolve (SPMAT       *A, PERM        *pivot,
				    VEC	*b, VEC *x);
extern "C" double _v_norm2   (VEC *x  ,VEC *scale);
extern "C" VEC *v_get(int);

/**@name matrix.C */
//@{

// sets the entries corresponding to folded and unfolded set to 1 and 0; the others to a negative value. assumes that mapping[i] == inf if and only if i is in folded or unfolded set.
void  computeInitialPfoldVector(vector<Real> & retval, int numPfoldsRequested,
				const vector<int>  & mapping, 
				const set<int> & foldedIndicesSet, int inf)
{
  retval.resize(numPfoldsRequested);
  for (int i = 0; i < numPfoldsRequested; i++)
    {
      if (mapping[i] == inf)
	{
	  if (foldedIndicesSet.find(i) != foldedIndicesSet.end())
	    retval[i] = 1;
	  else
	    retval[i] = 0;
	}
      else
	retval[i] = -1;
    }
}

/* 

   computes A, b, for pfold computation, and initial pfold vector.
   Returns the pfold initial vector, which only contains correct pfolds
   for nodes in the folded and unfolded set. The remaining pfolds are set to 
   a negative value.
   
   pseudocode:
     
     let total num nodes be n.
     
     let the set of nodes not in folded or unfolded state be C, (let there be m such nodes)
     the set of nodes in the folded state be F and the set of nodes in the 
     unfolded state be U.
  
     1. create a mapping from 1..n to 1..m, such that each node in C is mapped
     consecutively and other nodes are mapped to inf.


     2. for each node n_i in C:
          get its transition probabilities to its neighbours;
          normalize these transition probabilities so as it sums to 1 and the
          self transition probability is zero; (no need to do this here)
	  put the transition probabilities from n_i to nodes in C into T at the appropriate position (if the neighbor of n_i is n_j, put an entry into map(n_i), map(n_j)).
	  put the sum of transition probabilities from n_i to nodes in F into b.
	  (at location map(n_i)).

     3. Compute A = I - T;

     4. Compute return vector, of size numPfoldsRequested. 
     Use the mapping vector, as well as foldedIndicesSet for this purpose as follows:
     for each entry in mapping vector that points to inf, find out whether
     the pfold is equal to 1 or 0 (i.e. whether node folded or unfolded)
     for the remaining entries set return vector value to a negative value.
 */
vector<Real> getPfoldKnowns(SPMAT ** A, VEC ** b, Protein ** allNodes, 
			     int numNodes,
			     const vector<int>& foldedIndices,
			     const vector<int>& unfoldedIndices,
			     int maxNumNeighbors,
			     int numPfoldsRequested)
{
  SPMAT * T;
  vector<int>   mapping, neighborIndices;
  vector<float> tranProbs;
  vector<Real>  retval;
  const int   inf = -2208;
  float sumTransProb;
  set<int> foldedIndicesSet;
  int m;
  

  // 1. simple data structures precomputation
  computeMapping(mapping, foldedIndices, unfoldedIndices, numNodes, inf);
  assert (mapping.size() == (unsigned int) numNodes);
  
  for (unsigned int i = 0; i < foldedIndices.size(); i++)
    {
      foldedIndicesSet.insert(foldedIndices[i]);
    }

  //computeNotFoldedNotUnfoldedNodes(C, mapping, numNodes);
  //assert (C.size() == numNodes - foldedIndices.size() - unfoldedIndices.size());

  // 2. getting transition probabilities, setting elements of T and b.
  m  = numNodes - foldedIndices.size() - unfoldedIndices.size();
  T  = sp_get(m, m, maxNumNeighbors);
  *A = sp_get(m, m, maxNumNeighbors+1); // I-T may have one more nonzero in a row.
  *b = v_get(m); 
  // the elements of A and T and b should be 0 now.
  
  for (int i = 0; i < numNodes; i++)
    {
      if (mapping[i] == inf)
	{
	  // i in F or U.
	  continue;
	}
	
      findNeighborsAndTransitionProbabilities (neighborIndices,
					       tranProbs,
					       i, 
					       allNodes);
      assert (neighborIndices.size() == tranProbs.size());
      
      sumTransProb = 0.0;
      for (unsigned int j = 0; j < neighborIndices.size() ; j++)
	{
	  assert (tranProbs[j] >= 0);
	  sumTransProb      += tranProbs[j];
	}
      assert (sumTransProb > 0);

      sp_set_val(*A, mapping[i], mapping[i], 1.0);

      for (unsigned int j = 0; j < neighborIndices.size() ; j++)
	{

	  assert (neighborIndices[j] != i);
	  assert ((neighborIndices[j] >= 0) && 
		  (neighborIndices[j] < numNodes));
	  
	  float p_ij = tranProbs[j] / sumTransProb;
	  
	  if (mapping[neighborIndices[j]] != inf)
	    {
	      sp_set_val(T, mapping[i], mapping[neighborIndices[j]], 
			 p_ij);

	      sp_set_val(*A, mapping[i], mapping[neighborIndices[j]], 
			 -p_ij);
	    }
	  else if (foldedIndicesSet.find(neighborIndices[j]) != foldedIndicesSet.end())
	    {
	      (*b)->ve[mapping[i]] += p_ij;
	    }
	}
    }

  //3. matsub(A, I, T);
  // done above.
  
  //4. compute retval.
  computeInitialPfoldVector(retval, numPfoldsRequested, mapping, foldedIndicesSet, inf);
  return retval;
  
  
}

///
void computePfoldWrapper(vector<Real> & pfolds, 
			 int numNodes, Protein ** allNodes, 
			 int numPfoldsRequested, 
			 const vector<int>& foldedIndices,
			 const vector<int>& unfoldedIndices,
			 int maxNumNeighbors)
{
  

  SPMAT * A, *B;
  VEC   * b, *x, *v;
  PERM  * pivot;


  //vector<Real> pfolds;

  //cerr << "matrix.C: (using " << ((sizeof(Real)==sizeof(float))? "float": "double") << ")" << endl;
  pfolds = getPfoldKnowns(&A, &b, allNodes, numNodes, foldedIndices, 
			  unfoldedIndices, maxNumNeighbors, 
			  numPfoldsRequested);
  B = sp_copy(A);

  x = v_get(A->m);
  v = v_get(A->m);

  pivot = px_get(A->m);

  spLUfactor(A,pivot,0.1);
  spLUsolve (A,pivot,b,x);

  sp_mv_mlt (B,x,v);

  /* compute residual */
  v_sub(b,v,v);
  printf("relative residual : %g\n", v_norm2(v)/v_norm2(b));
  if ( v_norm2(v) >= MACHEPS*v_norm2(b)*A->m )
    {
      errmesg("spLUfactor()/spLUsolve()");
      printf("# Sparse LU residual = %g [cf MACHEPS = %g]\n",
	     v_norm2(v), MACHEPS);
      printf("threshold was : %g\n", MACHEPS*v_norm2(b)*A->m);
    }

  /* compute return value */
  cout << "pfolds size = " << pfolds.size() << " numpfoldsrequested = " << numPfoldsRequested << endl;
  assert (pfolds.size() == (unsigned int) numPfoldsRequested);
  int j = 0;
  for (int i = 0; i < numPfoldsRequested; i++)
    {
      if (pfolds[i] < 0)
	{
	  pfolds[i] = x->ve[j];
	  j++;
	}
    }
  /*  cout << "matrix.C" << endl;
  for (unsigned int i = 0; i < pfolds.size(); i++)
    {
      cout << pfolds[i] << endl;
    }
  */
  //return pfolds;


}

/* 
   creates a mapping from 0..n-1 to 0..m-1, where n = numNodes and m = numNodes
   *not* falling into the set of folded indices and unfolded indices.
   
   The mapping is a linearly increasing one.
*/
void computeMapping (vector<int> & mapping, const vector<int>& foldedIndices, 
		     const vector<int>& unfoldedIndices, int numNodes, 
		     int inf)
{
  unsigned int numSkipped;

  mapping.resize(numNodes);
  for (int i = 0; i < numNodes; i++)
    {
      mapping[i] = 0;
    }
  for (unsigned int i = 0; i < foldedIndices.size(); i++)
    {
      assert ((foldedIndices[i] >= 0) && (foldedIndices[i] < numNodes));
      mapping[foldedIndices[i]] = inf;
    }
  
  for (unsigned int i = 0; i < unfoldedIndices.size(); i++)
    {
      assert ((unfoldedIndices[i] >= 0) && (unfoldedIndices[i] < numNodes));
      mapping[unfoldedIndices[i]] = inf;
    }

  numSkipped = 0;
  for (int i = 0; i < numNodes; i++)
    {
      if (mapping[i] == inf)
	numSkipped++;
      else
	mapping[i] = i - numSkipped;
    }
  assert (numSkipped == foldedIndices.size() + unfoldedIndices.size());
}
//@}
