#include "roadmapConstruction.H"
#include <math.h>
#include "sse.H"
#include "nearestNeighbors.H"
#include <fstream>
#include "vec_utility.H"
#include "ioRoutines.H"
#include <algorithm>
using std::min;

//constants
const float  _2_M_PI         = 2 * M_PI;
const float UNINITIALIZED    = -2208.0;
const float  inv_kT          = 1.677481737;
const int    MAXNODES        = 1000000;

extern vector<int>     nn_idx;// near neighbor indices


//global variables
int           numTrees        = 0;
int           globalTreeName  = 0; //holds the id of the individual connected components.
treeClass *   globalFirstTree = NULL, //list of connected components is held in a linked list structure. this pointer points to the first connected component. 
  * globalLastTree = NULL; //this points to the last connected component.
Protein*      allNodes[MAXNODES];

/**@name roadmapConstruction.C */
//@{

/** 
   reads native structure, stores it in allNodes[endNodeIndex], 
   and calls necessary methods for initial setup. 
   also initializes pairOfVector structure, which holds pairwise vector 
   distances and angles.
   
   if hhbonds is set, reads the hhbond information.
*/

void initNativeStructure (int * numberOfNodes, const char * relDir, 
			  const char * protein_name, 
			  int endNodeIndex)
{
  nodeClass * newNode;
  vector<vector<float> > cACoords;
  char parsedDSSPFilename [100], parsedPDBFilename[100];
  vector<bool> cyclic;

  allNodes[endNodeIndex] = new Protein(endNodeIndex);
  strcpy( parsedDSSPFilename, relDir);
  strcat( parsedDSSPFilename, protein_name);
  strcpy( parsedPDBFilename , parsedDSSPFilename);
  strcat( parsedDSSPFilename, ".parsedDSSP");
  strcat( parsedPDBFilename , ".parsedPDB");
  
  allNodes[endNodeIndex] -> readParsedDSSPFile( parsedDSSPFilename );
  allNodes[endNodeIndex] -> readParsedPDBFile ( parsedPDBFilename , cACoords);

  updateSseSeq (Protein::sseTypeSeq);

  allNodes[endNodeIndex] -> computeSSEVectors (cACoords);
  
  allNodes[endNodeIndex] -> doSetup  (protein_name);
  
  newNode                 = new nodeClass;
  newNode->myConfig       = allNodes  [endNodeIndex];
  allNodes[endNodeIndex] -> setMyNode (newNode); 

  (* numberOfNodes )      = 1;


  standardizeStructure(allNodes[endNodeIndex]);

  // regenerating these structures since the coordinates read are in
  // arbitrary reference frame.
  allNodes[endNodeIndex]->generateCentroidCoords();
  allNodes[endNodeIndex]->generateCAlphaCoords();
  allNodes[endNodeIndex]->findParams();
  allNodes[endNodeIndex]->setPairwiseDistances();
  allNodes[endNodeIndex]->_computeEnergy();

  getParameterInfo                    (cyclic);
  
  Protein::cyclic.resize(0);

  for (unsigned int i = 0; i < cyclic.size(); i++)
    { 
      cout << i << " th parameter is :";
      Protein::cyclic.push_back(cyclic[i]);
      if (Protein::cyclic[i])
	{
	  cout << " cyclic" << endl;
	}
      else
	{
	  cout << " not cyclic" << endl;
	}
	
    }


}

/**
  creates the extended structure. 
*/
void initExtendedStructure(int * numberOfNodes, Protein * nativeP)
{
  /*
    the      parameters are:
    vector   angle : 180.
    dihedral angle : 0.
    twistangle     : 0.
    length         : as in folded state.
   */
  vector<float> dih, vec, twi, len;
  int           numVectors;

  numVectors = Protein::getNumSseElements();

  for (int i = 0; i < numVectors; i++)
    {
      dih.push_back(0.0); vec.push_back(M_PI); twi.push_back(0.0);
      float length = nativeP->getSseElement(i)->getLength();
      len.push_back(length);
    }
  
  allNodes[*numberOfNodes] = new Protein(*numberOfNodes);

  nodeClass * newNode                  = new nodeClass;
  newNode->myConfig                    = allNodes  [*numberOfNodes];

  allNodes[*numberOfNodes] -> createGivenConfig(dih, vec, twi, len, nativeP);

  allNodes[*numberOfNodes] -> setMyNode             (newNode); 
  allNodes[*numberOfNodes] -> findParams            ();
  allNodes[*numberOfNodes] -> computeEnergy        ();

  //cout << "extended node is : " << endl;
  //cout << (*allNodes[*numberOfNodes]) << endl;

  (*numberOfNodes) ++;
}

///
void kNNLinkNode(int nodeIndex, int numNodes, Protein ** allNodes, int k)
{

  findNearestNeighbors(nodeIndex, numNodes, allNodes, k);

  for (int i = 0; (i <  k) && 
	 (allNodes[nodeIndex]->getNumLinks() < k);   
       i++) 
    {
      int otherNodeIndex  = nn_idx[i];

      if (!connected(nodeIndex, otherNodeIndex,allNodes, numNodes))
	{
	  makeLinkBetweenNodes (numNodes, nodeIndex,otherNodeIndex,
				UNINITIALIZED, 
				UNINITIALIZED);
	  /*
	    float d = distance(nodeIndex, nn_idx[i]);
	    //assuming distance is symmetric
	    assert (d == distance(nn_idx[i], nodeIndex);
	    if (distance(nodeIndex, nn_idx[i])
	  */
	}
    }
	  
  if (allNodes[nodeIndex]->myTree == NULL)
    addNode2Forest(nodeIndex);
}



/**
  connects each node to k-NN 
  
  returns the max. number of neighbors of any node.

*/
void linkNodes(int numNodes, int runIndex, int k)
{

  int nodeIndex;//, tempCreated = 0, tempDestroyed = 0;
  
  initializekNN (numNodes, k, 1);
  
  for (nodeIndex = 0; nodeIndex< numNodes; nodeIndex++) 
    {
      if ((nodeIndex % 500) == 0) 
	cout << "nodeIndex =" << nodeIndex << endl;
      
      if (allNodes[nodeIndex]->getNumLinks() < k)
	{

	  kNNLinkNode(nodeIndex, numNodes, allNodes, k);

	}
    }
  linkNodesStatPrintOuts( numNodes,  numTrees, allNodes, globalFirstTree);
  //  cout << "Trees created " << tempCreated << " trees removed " << tempDestroyed << endl;
  

}



/** connects n1 and n2 with the given pathweights. 

    returns the value returned by the call to joinTheTreesForNodes function.

    does not touch transition probabilities, which are computed later. 
*/
int makeLinkBetweenNodes (int numNodes, const int n1, const int n2, 
			  const float forwardPathWeight, 
			  const float backwardPathWeight) 
{
  int retval;

  makeOneWayLink ( n1,  n2,   forwardPathWeight);
  makeOneWayLink ( n2,  n1,   backwardPathWeight);
  
  retval = joinTheTreesForNodes(n1, n2, numNodes);

  return retval;
}

/**
   does join the trees. I don't know what it returns, but 
   I don't think it is used.
*/
int joinTheTreesForNodes (int n1, int n2, int numNodes)
{
  treeClass * tempTreePtr;
  int retval = 0;

  if ((allNodes[n2]->myTree == NULL) && (allNodes[n1]->myTree == NULL)) 
    {
      //    cout << " A new tree is created. ";
      retval                      = 1;
      
      tempTreePtr                 = new treeClass;
      
      tempTreePtr->root           = allNodes[n2]->myNode;
      tempTreePtr->numNodesInTree = 2;
      tempTreePtr->prevTree       = globalLastTree;
      tempTreePtr->myTreeName     = globalTreeName;
      globalTreeName++;
      
      allNodes[n2]->myTree        = tempTreePtr;
      allNodes[n1]->myTree        = tempTreePtr;
      
      if (globalLastTree == NULL) 
	{
	  globalLastTree  = tempTreePtr;
	  globalFirstTree = tempTreePtr;
	}
      else 
	{
	  globalLastTree->nextTree = tempTreePtr;
	  globalLastTree           = tempTreePtr;
	}
      numTrees++;
    }
  else if ((allNodes[n2]->myTree == NULL) && (allNodes[n1]->myTree != NULL)) 
    {
      allNodes[n2]->myTree = allNodes[n1]->myTree;
      allNodes[n1]->myTree->numNodesInTree++;
    }
  else if ((allNodes[n2]->myTree != NULL) && (allNodes[n1]->myTree == NULL))
    {
      allNodes[n1]->myTree = allNodes[n2]->myTree;
      allNodes[n2]->myTree->numNodesInTree++;
    }
  else if ((allNodes[n2]->myTree != NULL) && 
	   (allNodes[n1]->myTree != NULL) && 
	   (allNodes[n2]->myTree != allNodes[n1]->myTree))
    {
      tempTreePtr = allNodes[n2]->myTree;
      tempTreePtr->removeTree (&numTrees, allNodes[n1]->myTree, 
			       &globalFirstTree, &globalLastTree, allNodes, 
			       numNodes);
      delete tempTreePtr;
      retval = -1;
    }
 
  
  return retval;
}


/** simply connects the node n1 to node n2. */
void makeOneWayLink (const int n1, const int n2, const float pW)
{ 
  assert (allNodes[n1] -> myNode != NULL);
  assert (allNodes[n2] -> myNode != NULL);
  
  allNodes[n1]->myNode->addNodeLink(allNodes[n2]->myNode, pW);
  allNodes[n1]->numLinks++;
  

}


/**
  adds allNodes[count] to the global forest.
  the myTree parameter of allNodes[count] should be NULL.
*/

void addNode2Forest(int count)
{

  treeClass  *tempTreePtr;

  assert (allNodes[count]->myTree == NULL) ;
    
  //tempCreated++;
  
  tempTreePtr                 = new treeClass;
  tempTreePtr->root           = allNodes[count]->myNode;
  tempTreePtr->numNodesInTree = 1;
  tempTreePtr->prevTree       = globalLastTree;
  tempTreePtr->myTreeName     = globalTreeName;
  globalTreeName++;
	    
  allNodes[count]->myTree     = tempTreePtr;
	
  if (globalLastTree == NULL) 
    {
      globalLastTree = tempTreePtr;
      globalFirstTree = tempTreePtr;
    }
  else 
    {
      globalLastTree->nextTree = tempTreePtr;
      globalLastTree = tempTreePtr;
    }
	
  numTrees++;
	
   

}


/**
   creates trees for the nodes that are not connected to any other node.
*/
void   completeTreeGeneration(	int numNodes) 
{
  int i;
  treeClass * tempTreePtr;
  
  for (i = 0; i < numNodes; i++) {

    if (allNodes[i]->myTree == NULL) 
      {
	tempTreePtr                 = new treeClass;
	tempTreePtr->root           = allNodes[i]->myNode;
	tempTreePtr->numNodesInTree = 1;
	tempTreePtr->prevTree       = globalLastTree;
	tempTreePtr->myTreeName     = globalTreeName;
	globalTreeName++;
	    
	allNodes[i]->myTree         = tempTreePtr;
	
	if (globalLastTree == NULL) {
	  globalLastTree  = tempTreePtr;
	  globalFirstTree = tempTreePtr;
	}
	else {
	  globalLastTree->nextTree = tempTreePtr;
	  globalLastTree = tempTreePtr;
	}
	numTrees++;
      }
  }
  cout << "num trees = " << numTrees << endl;
}



/** 
    samples (numToCreate) conformations around node # origIndex.
    returns the index of the conformation who has the lowest energy
    in this process. This conformation may be the center node.
*/

int sampleAroundProtein (int * numNodes,  const int origIndex, 
			 const int numToCreate) 
			 
{
  
  int        numCreated  = 0;
  nodeClass *newNode;
  int        numCollided = 0;
  float      tempEnergy  = 0;

  int   minIndex    = origIndex;
  float minEnergy   = allNodes[minIndex]->getEnergy();


  while ((numCreated < numToCreate) && ((*numNodes) < MAXNODES)) 
    {
      allNodes[(*numNodes)] = new Protein(*numNodes);
      allNodes[(*numNodes)] -> createNonRandomConfig(allNodes[origIndex], 
						     allNodes);
    
      if  (allNodes[(*numNodes)]->getExclusionEnergy() < Protein::maxExEnergy)
	{
	  newNode                       = new nodeClass;
	  newNode->myConfig             = allNodes[(*numNodes)];
	  allNodes[(*numNodes)]->myNode = newNode;

	  tempEnergy                    = allNodes[(*numNodes)]->getEnergy();
	  
	  if (tempEnergy < minEnergy) 
	    {
	      minEnergy = tempEnergy;
	      minIndex  = (*numNodes);
	    }
	  
	  (*numNodes)++;
	  numCreated++;
	}
      else 
	{
	  numCollided++;
	  delete allNodes[(*numNodes)];
	}
    }
  return minIndex;
}

/**
   I take samples w/ parameters either gaussianly distributed around the origin 
   or randomly sampled values. 
   in gaussian sampling, when the 1/6th total num.  nodes to be sampled is sampled,
   the sigma coefficient is changed.
   
   now assumes that extended config. is also sampled bef. calling this routine.
//   samples all nodes around extended config.
*/

void createNodes (int * numberOfNodes, 
		  int numNodesAlreadySampled, int numNodes2Sample)
{
  
  int i, finalNumNodes, numAccepted = 0, numRejected = 0;
  float sigmaEnergy = 0.0;
  nodeClass * newNode;



  //  vector<float> params, newParams;
  //  int maxIter = 1000;
  //  bool nodeApart = false; // true if node distant from other nodes

  i             = numNodesAlreadySampled;
  finalNumNodes = numNodesAlreadySampled + numNodes2Sample;

  assert (finalNumNodes < MAXNODES);

   
  // assert (i == 1);
  assert (i > 1);

  //  params = allNodes[1]->getParams();
  
  while (i < finalNumNodes) 
    {
      bool foundANode = false; // becomes true if the node distance to 
      if (i % 500 == 0)
	cout << "total numnodes till now = " << i << endl;

      // other nodes is large enough 
      // int iter = 0; 
      // num times new sampled node is close enough to other nodes
      do 
	{
	  allNodes[i] = new Protein(i);

	  allNodes[i]->createRandomConfig(allNodes[0]);
	

	  if (allNodes[i]->getExclusionEnergy() >= Protein::maxExEnergy)
	    {
	    
	      delete allNodes[i];
	      numRejected ++;
	    }
	  else 
	    {
	      foundANode = true;
	      numAccepted ++;
	    }
	}
      while (!foundANode);// && (iter< maxIter));
      
      newNode                 = new nodeClass;
      newNode->myConfig       = allNodes[i];
      allNodes[i]  ->  setMyNode      (newNode);
      sigmaEnergy  +=  allNodes[i]->getEnergy();
      allNodes[i]  ->  checkBondDihedralAngles();
      //      cout << "printing created node : " << endl;
      //      cout << (*allNodes[i]);
      
      assert(allNodes[i]->getExclusionEnergy() < Protein::maxExEnergy);
      i++;
    }
    
  (*numberOfNodes) = finalNumNodes;
  cout << "node acceptance ratio in random sampling = " << numAccepted * 1.0/ (numAccepted + numRejected) << endl;
}

///
int mycompare(const void *vfirst, const void *vsecond) {

  neighbourType *first;
  neighbourType *second;
  
  first = (neighbourType*)(vfirst);
  second = (neighbourType*)(vsecond);

  if (first->dist < second->dist) {
    return -1;
  }
  else if (first->dist == second->dist) {
    return 0;
  }
  else {
    return 1;
  }
}

///assigns tr. prob. to the edge. Note: the assigned tr. prob. right now is different from the one in SRS jcb paper. I should change this.
void findTransitionProbability(int n1, int n2,float E1, float E2, 
			       float *fTP, float *bTP)

{
  float p = ((float)n1)/n2 * exp(-(E2 - E1) * inv_kT);
  
  assert (p != 0.0);
  *fTP = min<float>(1, p);
  *bTP = min<float>(1, 1/p);
}


///
void  assignTransitionProbability(int i,int j, float fTP, Protein ** allNodes)
{
  bool found;
  ptrList * linkPtr;

  linkPtr  = allNodes[i]->myNode->firstLink;
  found    = false;

  while ((linkPtr != NULL) && (!found))
    {
      int index = linkPtr->node->myConfig->getIndex();
      
      if (index == j)
	{
	  found = true;
	  assert (linkPtr->pathWeight == UNINITIALIZED);
	  linkPtr->edgeTransitionProbability = fTP;
	}
      linkPtr = linkPtr ->nextPtr;
    }
  assert (found);

}


/// returns maxNumNeighbors of any node.

unsigned int computeTransitionProbabilities(Protein ** allNodes, int numNodes)
{
  vector<int> neighbors;
  float fTP, bTP;
  unsigned int maxNumNeighbors = 0;

  for (int node1 = 0; node1 < numNodes; node1++)
    {
      findNeighbors (neighbors, node1, allNodes);

      if (neighbors.size() > maxNumNeighbors)
	maxNumNeighbors = neighbors.size();
      
      for (unsigned int k = 0; k < neighbors.size(); k++)
	{
	  int node2 = neighbors[k];
	  if (node2 > node1)
	    {
	      findTransitionProbability(allNodes[node1]->getNumLinks(),
					allNodes[node2]->getNumLinks(),
					allNodes[node1]->getEnergy(), 
					allNodes[node2]->getEnergy(), 
					&fTP, &bTP);
					  
	      assignTransitionProbability(node1,node2, fTP, allNodes);
	      assignTransitionProbability(node2,node1, bTP, allNodes);
	    }
	}
    }
  cout << "max. # of neighbors of a node is " << maxNumNeighbors << endl;
  return maxNumNeighbors;
}



/** 
    connects if they are not already connected.
*/
void connect(int nodeIndex, const vector<int> & neighbors, int numNodes)
{
  //cout << "connecting node # " << nodeIndex << " to : " << neighbors.size() << " neighbors." << endl;
  assert ((nodeIndex < numNodes) && (nodeIndex >= 0));

  for (unsigned int i = 0; i < neighbors.size(); i++)
    {
      assert ((neighbors[i] < numNodes) && (neighbors[i] >= 0));

      if (!connected(nodeIndex, neighbors[i], allNodes, numNodes))
	makeLinkBetweenNodes(numNodes, nodeIndex, neighbors[i],
			     UNINITIALIZED, UNINITIALIZED);
    }
}

void do_connection_steps(ofstream * outFile, int * numLinks, int count, int node2, const vector<float> & optimizedE, int numNodes)
{
  printEnergiesLine(outFile, count, node2, optimizedE);
		  
  makeLinkBetweenNodes (numNodes, count,node2,
			UNINITIALIZED, 
			UNINITIALIZED);
  (*numLinks) ++;
}



//samples m nodes around each of the centerNodeIndices.
void sampleAroundNodes(int * numberOfNodes, const vector<int> & centerNodeIndices, int m)
{
  for (unsigned int j   = 0; j < centerNodeIndices.size(); j++) 
    {
      sampleAroundProtein (numberOfNodes,  centerNodeIndices[j], m);
    }
}

//@}

